/*
 *	Name:		General Use Types Definitions -- Header Include file
 *	Version:	1.24
 *	Author:		Vladimir Arnost (QA-Software)
 *	Last revision:	Sep-4-1995
 *	Compiler:	Borland C++ 3.1, Watcom C/C++ 10.0
 *
 */

#ifndef __DEFTYPES_H_
#define __DEFTYPES_H_

/* Global type declarations */

#include "doomtype.h"

/* machine dependent types */
typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned int  uint;
typedef unsigned long ulong;

typedef signed char schar;
typedef signed short sshort;
typedef signed int  sint;
typedef signed long slong;

#endif // __DEFTYPES_H_
