// Microsoft Visual C++ generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "afxres.h"
#include "../version.h"

#include "../gitinfo.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE 
BEGIN
    "#include ""afxres.h""\r\n"
    "#include ""../version.h""\r\r\n"
    "#include ""../gitinfo.h""\r\0"
END

3 TEXTINCLUDE 
BEGIN
    "#ifndef NO_MANIFEST\r\n"
    "   CREATEPROCESS_MANIFEST_RESOURCE_ID RT_MANIFEST ""zdoom.exe.manifest""\r\n"
    "#endif\r\n"
    "\r\n"
    "/////////////////////////////////////////////////////////////////////////////\r\n"
    "//\r\n"
    "// Version\r\n"
    "//\r\n"
    "#define RC_FILEVERSION2 GIT_DESCRIPTION\r\n"
    "\r\n"
    "VS_VERSION_INFO VERSIONINFO\r\n"
    " FILEVERSION RC_FILEVERSION\r\n"
    " PRODUCTVERSION RC_PRODUCTVERSION\r\n"
    " FILEFLAGSMASK 0x3fL\r\n"
    "#ifdef _DEBUG\r\n"
    " FILEFLAGS 0x3L\r\n"
    "#else\r\n"
    " FILEFLAGS 0x2L\r\n"
    "#endif\r\n"
    " FILEOS 0x40004L\r\n"
    " FILETYPE 0x1L\r\n"
    " FILESUBTYPE 0x0L\r\n"
    "BEGIN\r\n"
    "    BLOCK ""StringFileInfo""\r\n"
    "    BEGIN\r\n"
    "        BLOCK ""040904b0""\r\n"
    "        BEGIN\r\n"
    "            VALUE ""Comments"", ""Thanks to id Software for creating DOOM and then releasing the source code. Thanks also to TeamTNT for creating BOOM, which ZDoom is partially based on. Includes code based on the Cajun Bot 0.97 by Martin Collberg.""\r\n"
    "            VALUE ""CompanyName"", "" ""\r\n"
    "            VALUE ""FileDescription"", ""ZDoom""\r\n"
    "            VALUE ""FileVersion"", RC_FILEVERSION2\r\n"
    "            VALUE ""InternalName"", ""ZDoom""\r\n"
    "            VALUE ""LegalCopyright"", ""Copyright \\u00A9 1993-1996 id Software, 1998-2010 Randy Heit, 2002-2010 Christoph Oelckers, et al.""\r\n"
    "            VALUE ""LegalTrademarks"", ""DoomR is a Registered Trademark of id Software, Inc.""\r\n"
    "            VALUE ""OriginalFilename"", ""zdoom.exe""\r\n"
    "            VALUE ""ProductName"", ""ZDoom""\r\n"
    "            VALUE ""ProductVersion"", RC_PRODUCTVERSION2\r\n"
    "        END\r\n"
    "    END\r\n"
    "    BLOCK ""VarFileInfo""\r\n"
    "    BEGIN\r\n"
    "        VALUE ""Translation"", 0x409, 1200\r\n"
    "    END\r\n"
    "EN\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_ICON1               ICON                    "icon1.ico"
IDI_BOING1              ICON                    "boing1.ico"
IDI_BOING2              ICON                    "boing2.ico"
IDI_BOING3              ICON                    "boing3.ico"
IDI_BOING4              ICON                    "boing4.ico"
IDI_BOING5              ICON                    "boing5.ico"
IDI_BOING6              ICON                    "boing6.ico"
IDI_BOING7              ICON                    "boing7.ico"
IDI_BOING8              ICON                    "boing8.ico"

/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO 
BEGIN
    IDD_IWADDIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 5
        RIGHTMARGIN, 205
        TOPMARGIN, 7
        BOTTOMMARGIN, 179
    END

    IDD_EAXPROPERTYLIST, DIALOG
    BEGIN
        LEFTMARGIN, 2
        RIGHTMARGIN, 249
        VERTGUIDE, 88
        VERTGUIDE, 122
        TOPMARGIN, 7
        BOTTOMMARGIN, 400
    END

    IDD_EAXEDIT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 273
        TOPMARGIN, 7
        BOTTOMMARGIN, 319
        HORZGUIDE, 299
    END

    IDD_NEWEAX, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 341
        TOPMARGIN, 7
        BOTTOMMARGIN, 250
        HORZGUIDE, 216
    END

    IDD_SAVEEAX, DIALOG
    BEGIN
        LEFTMARGIN, 6
        RIGHTMARGIN, 377
        VERTGUIDE, 370
        BOTTOMMARGIN, 88
        HORZGUIDE, 12
    END

    IDD_CRASHDIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 4
        RIGHTMARGIN, 408
        TOPMARGIN, 4
        BOTTOMMARGIN, 303
        HORZGUIDE, 49
        HORZGUIDE, 242
    END

    IDD_CRASHOVERVIEW, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 384
        TOPMARGIN, 7
        BOTTOMMARGIN, 257
    END

    IDD_CRASHDETAILS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 392
        TOPMARGIN, 7
        HORZGUIDE, 76
    END

    IDD_BOING, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 180
        VERTGUIDE, 33
        TOPMARGIN, 7
    END

    IDD_ERRORPANE, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 183
        TOPMARGIN, 7
        BOTTOMMARGIN, 21
    END

    IDD_NETSTARTPANE, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 182
        TOPMARGIN, 7
        BOTTOMMARGIN, 48
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_IWADDIALOG DIALOGEX 0, 0, 212, 186
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | DS_CENTER | WS_POPUP | WS_CAPTION | WS_SYSMENU
EXSTYLE WS_EX_APPWINDOW
CAPTION "Select an IWAD to use"
FONT 8, "MS Shell Dlg", 0, 0, 0x0
BEGIN
    LISTBOX         IDC_IWADLIST,5,25,200,124,LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    CONTROL         "Don't ask me this again",IDC_DONTASKIWAD,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,61,153,89,10
    DEFPUSHBUTTON   "OK",IDOK,5,165,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,155,165,50,14
    LTEXT           "ZDoom found more than one IWAD present.",IDC_STATIC,5,5,140,8
    LTEXT           "Select from the list below to determine which one to use:",IDC_STATIC,5,15,200,8
END

IDD_EAXPROPERTYLIST DIALOGEX 0, 0, 265, 404
STYLE DS_SETFONT | DS_3DLOOK | DS_FIXEDSYS | DS_CONTROL | WS_CHILD | WS_VSCROLL
EXSTYLE WS_EX_CLIENTEDGE
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    EDITTEXT        IDCE_ENVIRONMENTSIZE,76,7,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_ENVIRONMENTDIFFUSION,76,20,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_ROOM,76,33,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_ROOMHF,76,46,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_ROOMLF,76,59,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_DECAYTIME,76,72,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_DECAYHFRATIO,76,85,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_DECAYLFRATIO,76,98,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_REFLECTIONS,76,111,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_REFLECTIONSDELAY,76,124,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_REVERB,76,137,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_REVERBDELAY,76,150,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_ECHOTIME,76,163,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_ECHODEPTH,76,176,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_MODULATIONTIME,76,189,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_MODULATIONDEPTH,76,202,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_AIRABSORPTIONHF,76,215,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_HFREFERENCE,76,228,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_LFREFERENCE,76,241,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_ROOMROLLOFFFACTOR,76,254,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_REFLECTIONSPANX,76,267,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_REFLECTIONSPANY,76,280,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_REFLECTIONSPANZ,76,293,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_REVERBPANX,76,306,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_REVERBPANY,76,319,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    EDITTEXT        IDCE_REVERBPANZ,76,332,48,12,ES_AUTOHSCROLL,WS_EX_RIGHT
    CONTROL         "",IDCS_ENVIRONMENTSIZE,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,7,123,12
    CONTROL         "",IDCS_ENVIRONMENTDIFFUSION,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,20,123,12
    CONTROL         "",IDCS_ROOM,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,33,123,12
    CONTROL         "",IDCS_ROOMHF,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,46,123,12
    CONTROL         "",IDCS_ROOMLF,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,59,123,12
    CONTROL         "",IDCS_DECAYTIME,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,72,123,12
    CONTROL         "",IDCS_DECAYHFRATIO,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,85,123,12
    CONTROL         "",IDCS_DECAYLFRATIO,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,98,123,12
    CONTROL         "",IDCS_REFLECTIONS,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,111,123,12
    CONTROL         "",IDCS_REFLECTIONSDELAY,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,124,123,12
    CONTROL         "",IDCS_REVERB,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,137,123,12
    CONTROL         "",IDCS_REVERBDELAY,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,150,123,12
    CONTROL         "",IDCS_ECHOTIME,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,163,123,12
    CONTROL         "",IDCS_ECHODEPTH,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,176,123,12
    CONTROL         "",IDCS_MODULATIONTIME,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,189,123,12
    CONTROL         "",IDCS_MODULATIONDEPTH,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,202,123,12
    CONTROL         "",IDCS_AIRABSORPTIONHF,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,215,123,12
    CONTROL         "",IDCS_HFREFERENCE,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,228,123,12
    CONTROL         "",IDCS_LFREFERENCE,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,241,123,12
    CONTROL         "",IDCS_ROOMROLLOFFFACTOR,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,254,123,12
    CONTROL         "",IDCS_REFLECTIONSPANX,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,267,123,12
    CONTROL         "",IDCS_REFLECTIONSPANY,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,280,123,12
    CONTROL         "",IDCS_REFLECTIONSPANZ,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,293,123,12
    CONTROL         "",IDCS_REVERBPANX,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,306,123,12
    CONTROL         "",IDCS_REVERBPANY,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,319,123,12
    CONTROL         "",IDCS_REVERBPANZ,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,126,332,123,12
    CONTROL         "Reflections Scale",IDC_REFLECTIONSSCALE,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,19,353,70,10
    CONTROL         "Reflections Delay Scale",IDC_REFLECTIONSDELAYSCALE,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,19,366,90,10
    CONTROL         "Decay Time Scale",IDC_DECAYTIMESCALE,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,19,377,71,10
    CONTROL         "Decay HF Limit",IDC_DECAYHFLIMIT,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,19,390,63,10
    CONTROL         "Reverb Scale",IDC_REVERBSCALE,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,139,353,58,10
    CONTROL         "Reverb Delay Scale",IDC_REVERBDELAYSCALE,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,139,366,78,10
    CONTROL         "Echo Time Scale",IDC_ECHOTIMESCALE,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,139,377,67,10
    CONTROL         "Modulation Time Scale",IDC_MODULATIONTIMESCALE,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,139,390,86,10
    LTEXT           "Environment Size",IDC_STATIC,17,9,56,8,0,WS_EX_RIGHT
    LTEXT           "Environment Diffusion",IDC_STATIC,2,22,71,8,0,WS_EX_RIGHT
    LTEXT           "Room",IDC_STATIC,53,35,19,8,0,WS_EX_RIGHT
    LTEXT           "Room HF",IDC_STATIC,43,48,30,8,0,WS_EX_RIGHT
    LTEXT           "Room LF",IDC_STATIC,45,60,28,8,0,WS_EX_RIGHT
    LTEXT           "Decay Time",IDC_STATIC,35,74,38,8,0,WS_EX_RIGHT
    LTEXT           "Decay HF Ratio",IDC_STATIC,23,87,50,8,0,WS_EX_RIGHT
    LTEXT           "Decay LF Ratio",IDC_STATIC,24,100,49,8,0,WS_EX_RIGHT
    LTEXT           "Reflections",IDC_STATIC,37,113,36,8,0,WS_EX_RIGHT
    LTEXT           "Reflections Delay",IDC_STATIC,17,126,56,8,0,WS_EX_RIGHT
    LTEXT           "Reflections Pan X",IDC_STATIC,17,267,56,8,0,WS_EX_RIGHT
    LTEXT           "Reflections Pan Y",IDC_STATIC,17,280,56,8,0,WS_EX_RIGHT
    LTEXT           "Reflections Pan Z",IDC_STATIC,17,293,56,8,0,WS_EX_RIGHT
    LTEXT           "Reverb",IDC_STATIC,49,137,24,8,0,WS_EX_RIGHT
    LTEXT           "Reverb Delay",IDC_STATIC,29,150,44,8,0,WS_EX_RIGHT
    LTEXT           "Reverb Pan X",IDC_STATIC,29,306,44,8,0,WS_EX_RIGHT
    LTEXT           "Reverb Pan Y",IDC_STATIC,29,319,44,8,0,WS_EX_RIGHT
    LTEXT           "Reverb Pan Z",IDC_STATIC,29,332,44,8,0,WS_EX_RIGHT
    LTEXT           "Echo Time",IDC_STATIC,40,163,33,8,0,WS_EX_RIGHT
    LTEXT           "Echo Depth",IDC_STATIC,35,176,38,8,0,WS_EX_RIGHT
    LTEXT           "Modulation Time",IDC_STATIC,21,189,52,8,0,WS_EX_RIGHT
    LTEXT           "Modulation Depth",IDC_STATIC,16,202,57,8,0,WS_EX_RIGHT
    LTEXT           "Air Absorption HF",IDC_STATIC,16,215,57,8,0,WS_EX_RIGHT
    LTEXT           "HF Reference",IDC_STATIC,28,228,45,8,0,WS_EX_RIGHT
    LTEXT           "LF Reference",IDC_STATIC,29,241,44,8,0,WS_EX_RIGHT
    LTEXT           "Room Rolloff Factor",IDC_STATIC,9,254,64,8,0,WS_EX_RIGHT
END

IDD_EAXEDIT DIALOGEX 0, 0, 280, 326
STYLE DS_SETFONT | DS_FIXEDSYS | DS_CENTER | WS_POPUP | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME
EXSTYLE WS_EX_TOOLWINDOW | WS_EX_WINDOWEDGE
CAPTION "Environment Editor"
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    COMBOBOX        IDC_CURRENTENVIRONMENT,7,19,265,14,CBS_DROPDOWNLIST | CBS_SORT | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Static",IDC_DUMMY,7,36,265,263,0,WS_EX_CLIENTEDGE
    PUSHBUTTON      "&New...",IDC_NEW,8,305,45,14
    PUSHBUTTON      "&Save...",IDC_SAVE,60,305,45,14
    PUSHBUTTON      "&Revert",IDC_REVERT,112,305,45,14
    CONTROL         "&Test in level",IDC_TESTEAX,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,165,307,54,10
    PUSHBUTTON      "&Done",IDOK,228,305,45,14
    CONTROL         "Show &IDs in List",IDC_SHOWIDS,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,7,7,66,10
    LTEXT           "ID 2:",IDC_STATIC,239,7,19,8
    LTEXT           "ID 1:",IDC_STATIC,198,7,19,8
    LTEXT           "255",IDC_ID2,258,7,15,8
    LTEXT           "255",IDC_ID1,218,7,15,8
END

IDD_NEWEAX DIALOGEX 0, 0, 348, 257
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | DS_CENTERMOUSE | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Create New Environment..."
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    EDITTEXT        IDC_NEWENVNAME,32,216,187,12,ES_AUTOHSCROLL
    EDITTEXT        IDC_EDITID1,248,216,32,12,ES_AUTOHSCROLL | ES_NUMBER
    CONTROL         "",IDC_SPINID1,"msctls_updown32",UDS_SETBUDDYINT | UDS_ALIGNRIGHT | UDS_AUTOBUDDY | UDS_ARROWKEYS,236,216,11,14
    EDITTEXT        IDC_EDITID2,309,216,32,12,ES_AUTOHSCROLL | ES_NUMBER
    CONTROL         "",IDC_SPINID2,"msctls_updown32",UDS_SETBUDDYINT | UDS_ALIGNRIGHT | UDS_AUTOBUDDY | UDS_ARROWKEYS,297,216,11,14
    CONTROL         "",IDC_ENVIRONMENTLIST,"SysListView32",LVS_LIST | LVS_SINGLESEL | LVS_SHOWSELALWAYS | LVS_SORTASCENDING | LVS_ALIGNLEFT | WS_BORDER | WS_TABSTOP,7,18,334,192
    DEFPUSHBUTTON   "Create",IDOK,246,236,45,14
    PUSHBUTTON      "Cancel",IDCANCEL,296,236,45,14
    LTEXT           "Name:",IDC_STATIC,7,218,22,8
    LTEXT           "ID 1:",IDC_STATIC,226,218,17,8
    LTEXT           "ID 2:",IDC_STATIC,287,218,17,8
    LTEXT           "Based on:",IDC_STATIC,7,7,33,8
END

IDD_SAVEEAX DIALOGEX 0, 0, 394, 95
STYLE DS_SETFONT | DS_3DLOOK | DS_FIXEDSYS | DS_CONTROL | WS_CHILD | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_SYSMENU
EXSTYLE WS_EX_TRANSPARENT
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    GROUPBOX        "Include these environments:",IDC_SAVEGROUP,6,0,371,90,0,WS_EX_TRANSPARENT
    CONTROL         "",IDC_ENVLIST,"SysListView32",LVS_LIST | LVS_SHOWSELALWAYS | LVS_SORTASCENDING | LVS_ALIGNLEFT | WS_BORDER | WS_TABSTOP,12,12,303,72
    PUSHBUTTON      "Select All",IDC_SELECTALL,320,12,50,14
    PUSHBUTTON      "Select None",IDC_SELECTNONE,320,31,50,14
END

IDD_CRASHDIALOG DIALOGEX 0, 0, 415, 308
STYLE DS_SETFONT | DS_MODALFRAME | DS_SETFOREGROUND | DS_FIXEDSYS | DS_CENTER | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
EXSTYLE WS_EX_CONTROLPARENT | WS_EX_APPWINDOW
CAPTION "ZDoom Very Fatal Error"
FONT 8, "MS Shell Dlg", 400, 0, 0x0
BEGIN
    CONTROL         "",IDC_CRASHTAB,"SysTabControl32",WS_TABSTOP,4,4,404,280
    PUSHBUTTON      "&Send Error Report",IDYES,146,289,91,14,WS_DISABLED
    PUSHBUTTON      "Save Report to Disk...",IDC_SAVEREPORT,242,289,91,14
    PUSHBUTTON      "&Discard Report",IDNO,338,289,70,14
END

IDD_CRASHOVERVIEW DIALOGEX 1, 13, 400, 264
STYLE DS_SETFONT | DS_FIXEDSYS | DS_CONTROL | WS_CHILD | WS_VISIBLE
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    EDITTEXT        IDC_CRASHINFO,7,126,377,90,ES_MULTILINE | ES_AUTOVSCROLL | ES_WANTRETURN | WS_VSCROLL
    CONTROL         "",IDC_PLEASETELLUS,"RichEdit20A",ES_MULTILINE | ES_READONLY,7,50,377,54,WS_EX_TRANSPARENT
    LTEXT           "(Optional) To help diagnose the cause of this error, please describe what you were doing when the error occured. If you were playing with custom wads, please also indicate where to obtain them:",IDC_STATIC,7,108,377,16,0,WS_EX_TRANSPARENT
    CONTROL         131,IDC_DEADGUYVIEWER,"Static",SS_BITMAP,345,7,23,26
    CONTROL         "",IDC_CRASHHEADER,"RichEdit20A",ES_MULTILINE | ES_AUTOHSCROLL | ES_READONLY,18,13,306,25,WS_EX_TRANSPARENT
    CONTROL         "",IDC_STATIC,"Static",SS_ETCHEDHORZ,0,42,407,1
    GROUPBOX        "Summary",IDC_STATIC,7,223,377,34
    LTEXT           "Static",IDC_CRASHSUMMARY,14,233,363,20
END

IDD_CRASHDETAILS DIALOGEX 0, 0, 400, 164
STYLE DS_SETFONT | DS_3DLOOK | DS_FIXEDSYS | DS_CONTROL | WS_CHILD
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    LTEXT           "The error report contains these files:",IDC_STATIC,7,5,119,8
    LTEXT           "The selected file contains this information:",IDC_STATIC,7,74,136,8
    RTEXT           "Static",IDC_CRASHFILESIZE,337,74,55,8
    LISTBOX         IDC_CRASHFILES,7,15,385,53,LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    CONTROL         "",IDC_CRASHFILECONTENTS,"RichEdit20A",ES_MULTILINE | ES_READONLY | WS_BORDER | WS_VSCROLL | WS_TABSTOP,7,83,385,174
END

IDD_BOING DIALOGEX 0, 0, 187, 196
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | DS_CENTERMOUSE | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Upload status"
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "Close",IDOK,130,175,50,14,WS_DISABLED
    ICON            IDI_BOING1,IDC_BOING,7,7,21,20
    LTEXT           "Static",IDC_BOINGSTATUS,33,12,147,8
    CONTROL         "",IDC_STATIC,"Static",SS_ETCHEDHORZ,0,166,189,1
    CONTROL         "",IDC_BOINGEDIT,"RichEdit20A",ES_MULTILINE | ES_READONLY | WS_BORDER | WS_VSCROLL | WS_TABSTOP,7,34,173,124
    CONTROL         "",IDC_BOINGPROGRESS,"msctls_progress32",NOT WS_VISIBLE | WS_BORDER | 0x1,33,22,147,9
END

IDD_ERRORPANE DIALOGEX 0, 0, 190, 28
STYLE DS_SETFONT | DS_FIXEDSYS | WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN
EXSTYLE WS_EX_CONTROLPARENT
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "Quit",IDOK,133,7,50,14
END

IDD_NETSTARTPANE DIALOGEX 0, 0, 189, 55
STYLE DS_SETFONT | DS_3DLOOK | DS_FIXEDSYS | WS_CHILD | WS_CLIPSIBLINGS | WS_CLIPCHILDREN
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    PUSHBUTTON      "Abort Network Game",IDCANCEL,45,34,89,14
    CONTROL         "",IDC_NETSTARTMESSAGE,"Static",SS_LEFTNOWORDWRAP | SS_NOPREFIX | WS_GROUP,7,7,116,8
    CONTROL         "",IDC_NETSTARTCOUNT,"Static",SS_LEFTNOWORDWRAP | SS_NOPREFIX | WS_GROUP,153,7,29,8,WS_EX_RIGHT
    CONTROL         "",IDC_NETSTARTPROGRESS,"msctls_progress32",WS_BORDER,7,18,175,10
END


/////////////////////////////////////////////////////////////////////////////
//
// Bitmap
//

IDB_DEADGUY             BITMAP                  "deadguy.bmp"
#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//
#ifndef NO_MANIFEST
   CREATEPROCESS_MANIFEST_RESOURCE_ID RT_MANIFEST "zdoom.exe.manifest"
#endif

/////////////////////////////////////////////////////////////////////////////
//
// Version
//
#define RC_FILEVERSION2 GIT_DESCRIPTION

VS_VERSION_INFO VERSIONINFO
 FILEVERSION RC_FILEVERSION
 PRODUCTVERSION RC_PRODUCTVERSION
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x3L
#else
 FILEFLAGS 0x2L
#endif
 FILEOS 0x40004L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "Comments", "Thanks to id Software for creating DOOM and then releasing the source code. Thanks also to TeamTNT for creating BOOM, which ZDoom is partially based on. Includes code based on the Cajun Bot 0.97 by Martin Collberg."
            VALUE "CompanyName", " "
            VALUE "FileDescription", "ZDoom"
            VALUE "FileVersion", RC_FILEVERSION2
            VALUE "InternalName", "ZDoom"
            VALUE "LegalCopyright", "Copyright \u00A9 1993-1996 id Software, 1998-2010 Randy Heit, 2002-2010 Christoph Oelckers, et al."
            VALUE "LegalTrademarks", "DoomR is a Registered Trademark of id Software, Inc."
            VALUE "OriginalFilename", "zdoom.exe"
            VALUE "ProductName", "ZDoom"
            VALUE "ProductVersion", RC_PRODUCTVERSION2
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END
/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

