#ifndef __EXCEPT_H
#define __EXCEPT_H

#ifdef _MSC_VER
//==========================================================================
//
// CheckException
//
//==========================================================================

#ifndef FACILITY_VISUALCPP
#define FACILITY_VISUALCPP  ((LONG)0x6d)
#endif
#define VcppException(sev,err)  ((sev) | (FACILITY_VISUALCPP<<16) | err)

inline int CheckException(DWORD code)
{
	if (code == VcppException(ERROR_SEVERITY_ERROR,ERROR_MOD_NOT_FOUND) ||
		code == VcppException(ERROR_SEVERITY_ERROR,ERROR_PROC_NOT_FOUND))
	{
		return EXCEPTION_EXECUTE_HANDLER;
	}
	return EXCEPTION_CONTINUE_SEARCH;
}


#else

#ifdef __try
#undef __try
#endif
#define __try

#ifdef __except
#undef __except
#endif
#define __except(a) if (0)

#endif

#endif