/* config.h.in.  Generated from configure.in by autoheader.  */

/* Define to 1 if you have the `strdup' function. */
#define HAVE_STRDUP

/* Define to 1 if you have the `strndup' function. */
#define HAVE_STRNDUP

/* Name of package */
/* #undef PACKAGE */

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT "re2c-general@lists.sourceforge.net"

/* Define to the full name of this package. */
#define PACKAGE_NAME "re2c"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "re2c 0.12.3"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "re2c"

/* Define to the version of this package. */
#define PACKAGE_VERSION "0.12.3"

/* The size of `char', as computed by sizeof. */
#define SIZEOF_CHAR 1

/* The size of `int', as computed by sizeof. */
#define SIZEOF_INT 4

/* The size of `long', as computed by sizeof. */
#define SIZEOF_LONG 8

/* The size of `short', as computed by sizeof. */
#define SIZEOF_SHORT 2

/* Version number of package */
#define VERSION "0.12.3"

