xx(TK_Identifier,			"identifier")
xx(TK_StringConst,			"string constant")
xx(TK_NameConst,			"name constant")
xx(TK_IntConst,				"integer constant")
xx(TK_FloatConst,			"float constant")
xx(TK_DotDot,				"'..'")
xx(TK_Ellipsis,				"'...'")
xx(TK_RShiftEq,				"'>>='")
xx(TK_URShiftEq,			"'>>>='")
xx(TK_LShiftEq,				"'<<='")
xx(TK_AddEq,				"'+='")
xx(TK_SubEq,				"'-='")
xx(TK_MulEq,				"'*='")
xx(TK_DivEq,				"'/='")
xx(TK_ModEq,				"'%='")
xx(TK_AndEq,				"'&='")
xx(TK_XorEq,				"'^='")
xx(TK_OrEq,					"'|='")
xx(TK_RShift,				"'>>'")
xx(TK_URShift,				"'>>>'")
xx(TK_LShift,				"'<<'")
xx(TK_Incr,					"'++'")
xx(TK_Decr,					"'--'")
xx(TK_AndAnd,				"'&&'")
xx(TK_OrOr,					"'||'")
xx(TK_Leq,					"'<='")
xx(TK_Geq,					"'>='")
xx(TK_Eq,					"'=='")
xx(TK_Neq,					"'!='")
xx(TK_ApproxEq,				"'~=='")
xx(TK_LtGtEq,				"'<>='")
xx(TK_MulMul,				"'**'")
xx(TK_Action,				"'action'")
xx(TK_Break,				"'break'")
xx(TK_Case,					"'case'")
xx(TK_Const,				"'const'")
xx(TK_Continue,				"'continue'")
xx(TK_Default,				"'default'")
xx(TK_Do,					"'do'")
xx(TK_Else,					"'else'")
xx(TK_For,					"'for'")
xx(TK_If,					"'if'")
xx(TK_Return,				"'return'")
xx(TK_Switch,				"'switch'")
xx(TK_Until,				"'until'")
xx(TK_While,				"'while'")
xx(TK_Bool,					"'bool'")
xx(TK_Float,				"'float'")
xx(TK_Double,				"'double'")
xx(TK_Char,					"'char'")
xx(TK_Byte,					"'byte'")
xx(TK_SByte,				"'sbyte'")
xx(TK_Short,				"'short'")
xx(TK_UShort,				"'ushort'")
xx(TK_Int,					"'int'")
xx(TK_UInt,					"'uint'")
xx(TK_Long,					"'long'")
xx(TK_ULong,				"'ulong'")
xx(TK_Void,					"'void'")
xx(TK_Struct,				"'struct'")
xx(TK_Class,				"'class'")
xx(TK_Mode,					"'mode'")
xx(TK_Enum,					"'enum'")
xx(TK_Name,					"'name'")
xx(TK_String,				"'string'")
xx(TK_Sound,				"'sound'")
xx(TK_State,				"'state'")
xx(TK_Color,				"'color'")
xx(TK_Goto,					"'goto'")
xx(TK_Abstract,				"'abstract'")
xx(TK_ForEach,				"'foreach'")
xx(TK_True,					"'true'")
xx(TK_False,				"'false'")
xx(TK_None,					"'none'")
xx(TK_New,					"'new'")
xx(TK_InstanceOf,			"'instanceof'")
xx(TK_Auto,					"'auto'")
xx(TK_Exec,					"'exec'")
xx(TK_DefaultProperties,	"'defaultproperties'")
xx(TK_Native,				"'native'")
xx(TK_Var,					"'var'")
xx(TK_Out,					"'out'")
xx(TK_Ref,					"'ref'")
xx(TK_Event,				"'event'")
xx(TK_Static,				"'static'")
xx(TK_Transient,			"'transient'")
xx(TK_Volatile,				"'volatile'")
xx(TK_Final,				"'final'")
xx(TK_Throws,				"'throws'")
xx(TK_Extends,				"'extends'")
xx(TK_Public,				"'public'")
xx(TK_Protected,			"'protected'")
xx(TK_Private,				"'private'")
xx(TK_Dot,					"'dot'")
xx(TK_Cross,				"'cross'")
xx(TK_Ignores,				"'ignores'")
xx(TK_Localized,			"'localized'")
xx(TK_Latent,				"'latent'")
xx(TK_Singular,				"'singular'")
xx(TK_Config,				"'config'")
xx(TK_Coerce,				"'coerce'")
xx(TK_Iterator,				"'iterator'")
xx(TK_Optional,				"'optional'")
xx(TK_Export,				"'expert'")
xx(TK_Virtual,				"'virtual'")
xx(TK_Super,				"'super'")
xx(TK_Global,				"'global'")
xx(TK_Self,					"'self'")
xx(TK_Stop,					"'stop'")
xx(TK_Include,				"'include'")
xx(TK_Fixed_t,				"'fixed_t'")
xx(TK_Angle_t,				"'angle_t'")
xx(TK_Abs,					"'abs'")
xx(TK_Random,				"'random'")
xx(TK_Random2,				"'random2'")
xx(TK_FRandom,				"'frandom'")
xx(TK_Is,					"'is'")
xx(TK_Replaces,				"'replaces'")
xx(TK_Vector,				"'vector'")
xx(TK_Map,					"'map'")
xx(TK_Array,				"'array'")
xx(TK_In,					"'in'")
xx(TK_SizeOf,				"'sizeof'")
xx(TK_AlignOf,				"'alignof'")
xx(TK_RandomPick,			"'randompick'")
xx(TK_FRandomPick,			"'frandompick'")
#undef xx
