//{{NO_DEPENDENCIES}}
// Microsoft Visual C++ generated include file.
// Used by zdoom.rc
//
#define IDC_SAVEREPORT                  8
#define IDNO2                           9
#define IDI_ICON1                       101
#define IDD_MIDASINITERROR              108
#define IDD_IWADDIALOG                  112
#define IDC_INVISIBLECURSOR             114
#define IDD_CRASHDIALOG                 122
#define IDD_EAXPROPERTYLIST             123
#define IDD_DIALOG1                     124
#define IDD_EAXEDIT                     124
#define IDD_DIALOG2                     125
#define IDD_NEWEAX                      125
#define IDD_DIALOG3                     127
#define IDD_DIALOG4                     128
#define IDD_SAVEEAX                     128
#define IDB_BITMAP1                     131
#define IDB_DEADGUY                     131
#define IDD_CRASHDETAILS                133
#define IDI_BOING1                      137
#define IDI_BOING2                      138
#define IDI_BOING3                      139
#define IDI_BOING4                      140
#define IDI_BOING5                      141
#define IDI_BOING6                      142
#define IDI_BOING7                      143
#define IDI_BOING8                      144
#define IDD_BOING                       145
#define IDD_CRASHOVERVIEW               147
#define IDD_ERRORPANE                   148
#define IDD_NETSTARTPANE                149
#define IDC_ERRORMESSAGE                1004
#define IDQUIT                          1005
#define IDC_IWADLIST                    1006
#define IDC_DONTASKIWAD                 1007
#define IDC_SPIN1                       1010
#define IDC_SPINID1                     1010
#define IDC_CRASHINFO                   1011
#define IDC_SPINID2                     1011
#define IDC_TOOLTIP1                    1012
#define IDCE_ENVIRONMENTSIZE            1025
#define IDCS_ENVIRONMENTSIZE            1026
#define IDC_DECAYTIMESCALE              1027
#define IDC_REFLECTIONSSCALE            1028
#define IDC_REFLECTIONSDELAYSCALE       1029
#define IDC_REVERBSCALE                 1030
#define IDC_REVERBDELAYSCALE            1031
#define IDC_DECAYHFLIMIT                1032
#define IDC_ECHOTIMESCALE               1033
#define IDC_MODULATIONTIMESCALE         1034
#define IDC_DUMMY                       1035
#define IDC_NEW                         1036
#define IDC_SAVE                        1037
#define IDC_REVERT                      1038
#define IDC_CURRENTENVIRONMENT          1039
#define IDC_ID2                         1040
#define IDC_ID1                         1041
#define IDC_SHOWIDS                     1043
#define IDC_LIST3                       1047
#define IDC_ENVIRONMENTLIST             1047
#define IDC_SIZEBOX                     1048
#define IDC_EDIT1                       1049
#define IDC_NEWENVNAME                  1049
#define IDC_CRASHFILECONTENTS           1049
#define IDC_BOINGEDIT                   1049
#define IDC_EDITID1                     1050
#define IDC_EDITID2                     1051
#define IDC_HOTKEY1                     1053
#define IDC_LIST1                       1054
#define IDC_ENVLIST                     1054
#define IDC_CRASHFILES                  1054
#define IDC_SELECTALL                   1055
#define IDC_SELECTNONE                  1056
#define IDC_SAVEGROUP                   1057
#define IDC_TESTEAX                     1058
#define IDC_RICHEDIT21                  1059
#define IDC_CRASHHEADER                 1059
#define IDC_RICHEDIT22                  1061
#define IDC_PLEASETELLUS                1061
#define IDC_BUTTON2                     1062
#define IDC_CRASHDETAILS                1062
#define IDC_DEADGUYVIEWER               1063
#define IDC_BOING                       1065
#define IDC_CRASHFILESIZE               1066
#define IDC_BUTTON1                     1071
#define IDC_BOINGSTATUS                 1072
#define IDC_BOINGPROGRESS               1073
#define IDC_TAB1                        1074
#define IDC_CRASHTAB                    1074
#define IDC_RICHEDIT23                  1075
#define IDC_CRASHSUMMARY                1075
#define IDC_ERRORTEXT                   1076
#define IDC_ICONPIC                     1077
#define IDC_DIVIDERBAR                  1078
#define IDC_PROGRESS1                   1079
#define IDC_NETSTARTPROGRESS            1079
#define IDC_NETSTARTMESSAGE             1080
#define IDC_NETSTARTCOUNT               1081
#define IDC_STATIC_TITLE				1082
#define IDC_STATIC_STARTUP				1083
#define IDCE_ENVIRONMENTDIFFUSION       1085
#define IDCS_ENVIRONMENTDIFFUSION       1086
#define IDCE_ROOM                       1087
#define IDCS_ROOM                       1088
#define IDCE_ROOMHF                     1089
#define IDCS_ROOMHF                     1090
#define IDCE_ROOMLF                     1091
#define IDCS_ROOMLF                     1092
#define IDCE_DECAYTIME                  1093
#define IDCS_DECAYTIME                  1094
#define IDCE_DECAYHFRATIO               1095
#define IDCS_DECAYHFRATIO               1096
#define IDCE_DECAYLFRATIO               1097
#define IDCS_DECAYLFRATIO               1098
#define IDCE_REFLECTIONS                1099
#define IDCS_REFLECTIONS                1100
#define IDCE_REFLECTIONSDELAY           1101
#define IDCS_REFLECTIONSDELAY           1102
#define IDCE_REFLECTIONSPANX            1103
#define IDCS_REFLECTIONSPANX            1104
#define IDCE_REFLECTIONSPANY            1105
#define IDCS_REFLECTIONSPANY            1106
#define IDCE_REFLECTIONSPANZ            1107
#define IDCS_REFLECTIONSPANZ            1108
#define IDCE_REVERB                     1109
#define IDCS_REVERB                     1110
#define IDCE_REVERBDELAY                1111
#define IDCS_REVERBDELAY                1112
#define IDCE_REVERBPANX                 1113
#define IDCS_REVERBPANX                 1114
#define IDCE_REVERBPANY                 1115
#define IDCS_REVERBPANY                 1116
#define IDCE_REVERBPANZ                 1117
#define IDCS_REVERBPANZ                 1118
#define IDCE_ECHOTIME                   1119
#define IDCS_ECHOTIME                   1120
#define IDCE_ECHODEPTH                  1121
#define IDCS_ECHODEPTH                  1122
#define IDCE_MODULATIONTIME             1123
#define IDCS_MODULATIONTIME             1124
#define IDCE_MODULATIONDEPTH            1125
#define IDCS_MODULATIONDEPTH            1126
#define IDCE_AIRABSORPTIONHF            1127
#define IDCS_AIRABSORPTIONHF            1128
#define IDCE_HFREFERENCE                1129
#define IDCS_HFREFERENCE                1130
#define IDCE_LFREFERENCE                1131
#define IDCS_LFREFERENCE                1132
#define IDCE_ROOMROLLOFFFACTOR          1133
#define IDCS_ROOMROLLOFFFACTOR          1134
#define IDCS_REFLECTIONSDELAY2          1135
#define IDCE_MODULATIONTIME2            1135

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        150
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         1084
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif
