#define XLAT_NOP                             1
#define XLAT_OR                              2
#define XLAT_XOR                             3
#define XLAT_AND                             4
#define XLAT_MINUS                           5
#define XLAT_PLUS                            6
#define XLAT_MULTIPLY                        7
#define XLAT_DIVIDE                          8
#define XLAT_MODULUS                         9
#define XLAT_NEG                            10
#define XLAT_NUM                            11
#define XLAT_LPAREN                         12
#define XLAT_RPAREN                         13
#define XLAT_DEFINE                         14
#define XLAT_SYM                            15
#define XLAT_RBRACE                         16
#define XLAT_ENUM                           17
#define XLAT_LBRACE                         18
#define XLAT_COMMA                          19
#define XLAT_EQUALS                         20
#define XLAT_TAG                            21
#define XLAT_LBRACKET                       22
#define XLAT_RBRACKET                       23
#define XLAT_FLAGS                          24
#define XLAT_ARG2                           25
#define XLAT_ARG3                           26
#define XLAT_ARG4                           27
#define XLAT_ARG5                           28
#define XLAT_OR_EQUAL                       29
#define XLAT_COLON                          30
#define XLAT_MAXLINESPECIAL                 31
#define XLAT_SEMICOLON                      32
#define XLAT_SECTOR                         33
#define XLAT_NOBITMASK                      34
#define XLAT_BITMASK                        35
#define XLAT_CLEAR                          36
#define XLAT_LSHASSIGN                      37
#define XLAT_RSHASSIGN                      38
#define XLAT_LINEFLAG                       39
