// Copyright (c) 2006, 2007 Julio M. Merino Vidal
// Copyright (c) 2008 Ilya Sokolov, Boris Schaeling
// Copyright (c) 2009 Boris Schaeling
// Copyright (c) 2010 Felipe Tanus, Boris Schaeling
// Copyright (c) 2011, 2012 Jeff Flinn, Boris Schaeling
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

/**
 * \file boost/process/all.hpp
 *
 * Convenience header which includes all public Boost.Process header files.
 */

#ifndef BOOST_PROCESS_ALL_HPP
#define BOOST_PROCESS_ALL_HPP

#include <boost/process/child.hpp>
#include <boost/process/create_pipe.hpp>
#include <boost/process/execute.hpp>
#include <boost/process/executor.hpp>
#include <boost/process/initializers.hpp>
#include <boost/process/pipe.hpp>
#include <boost/process/search_path.hpp>
#include <boost/process/shell_path.hpp>
#include <boost/process/terminate.hpp>
#include <boost/process/wait_for_exit.hpp>

#endif
